/*****************************************************************************/
/*                      WiNRADiO G65DDC EXCALIBUR SIGMA                      */
/*                       Coherent mode SDK DEMO - ETC                        */
/*                        (c) Copyright RADIXON 2023                         */
/*****************************************************************************/

#include <arpa/inet.h>
#include <sys/ioctl.h>
#include "u_rx_client.h"
#include <functional>

#define TIMEOUT 5000
#define MAX_DDC1_BUFFER_COUNT 20

rx_client::rx_client(uint32_t index,uint32_t device_count,int32_t delay,const std::string &host_name,in_addr addr,uint16_t port)
{
    m_socket=-1;
    m_addr=addr;
    m_host_name=host_name;
    m_port=port;
    m_connected=false;
    m_index=index;
    m_device_count=device_count;
    m_delay=delay;
    m_power=0;

    m_serial_number_changed=0;

    m_callback=NULL;

    m_address_str=inet_ntoa(addr);
}

rx_client::~rx_client(void)
{
    disconnect();
}

int rx_client::connect(void)
{
 sockaddr_in addr;
 int error;

    if(m_socket!=-1)
    {
        if(!m_connected)
        {
            return 0;
        }

        disconnect();
    }

    memset(&m_device_state,0,sizeof(m_device_state));

    for(auto& temp:m_device_state.Temperatures)
    {
        temp=G65DDC_TEMPERATURE_UNKNOWN;
    }

    m_1pps_ddc1_sample_counter=-1;
    m_1pps_event_counter=-1;


    m_socket=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);

    if(m_socket==-1)
    {
        return errno;
    }

    m_packet_seq=0;

    addr.sin_addr=m_addr;
    addr.sin_port=htons(m_port);
    addr.sin_family=AF_INET;

    if(::connect(m_socket,(sockaddr*)&addr,(int)sizeof(addr))==-1)
    {
        error=errno;
        close(m_socket);
        m_socket=-1;
        return error;
    }

    m_stop_request=false;

    m_connected=true;
    m_power=0;

    m_thread=new std::thread(std::bind(&rx_client::thread_procedure,this));

    send_extended_configuration_params();

    return 0;
}

void rx_client::disconnect(void)
{
    if(m_socket==-1)
    {
        return;
    }

    m_stop_request=true;

    m_thread->join();
    delete m_thread;

    m_thread=NULL;

    close(m_socket);

    m_socket=-1;
    m_connected=false;

    free_buffers();
}


void rx_client::thread_procedure(void)
{
 fd_set read_fd;
 int result;
 timeval timeout;
 union tcp_packet *packet;
 uint32_t bytes_received;
 int bytes_in_queue;
 uint32_t total_packet_size;
 uint32_t bytes_to_recv;

    packet=(union tcp_packet*)malloc(MAX_PACKET_SIZE);

    bytes_received=0;

    timeout.tv_sec=0;
    timeout.tv_usec=100000;

    while(!m_stop_request && m_connected)
    {
        FD_ZERO(&read_fd);
        FD_SET(m_socket,&read_fd);

        result=select(m_socket+1,&read_fd,NULL,NULL,&timeout);

        if(result>0)
        {
            if(ioctl(m_socket,FIONREAD,&bytes_in_queue)==0)
            {
                if(bytes_in_queue>0)
                {
                    do
                    {
                        if(bytes_received>=EMPTY_PACKET_SIZE)
                        {
                            total_packet_size=sizeof(packet_size_t)+packet->size;

                            bytes_to_recv=total_packet_size-bytes_received;

                            if(bytes_to_recv>(uint32_t)bytes_in_queue)
                            {
                                bytes_to_recv=bytes_in_queue;
                            }

                            result=recv(m_socket,packet->c+bytes_received,bytes_to_recv,0);

                            if(result>0)
                            {
                                bytes_received+=result;
                                bytes_in_queue-=result;

                                if(total_packet_size==bytes_received)
                                {
                                    process_packet(packet);
                                    bytes_received=0;
                                }
                            }
                            else if(result==0)
                            {
                                //disconnected
                                m_connected=false;
                                break;
                            }
                        }
                        else
                        {
                            bytes_to_recv=EMPTY_PACKET_SIZE-bytes_received;

                            if(bytes_to_recv>(uint32_t)bytes_in_queue)
                            {
                                bytes_to_recv=bytes_in_queue;
                            }

                            result=recv(m_socket,packet->c+bytes_received,bytes_to_recv,0);

                            if(result>0)
                            {
                                bytes_received+=result;
                                bytes_in_queue-=result;
                            }
                            else if(result==0)
                            {
                                //disconnected
                                m_connected=false;
                                break;
                            }
                        }
                    }
                    while(bytes_in_queue>0);
                }
                else
                {
                    m_connected=false;
                    break;
                }
            }
        }
    }

    m_cs.lock();
    m_serial_number.clear();
    m_cs.unlock();

    m_serial_number_changed=true;

    free(packet);
}



bool rx_client::send(union tcp_packet &packet,overlapped *o)
{
 int result;
 uint32_t bytes;
 const void *buffer;
 uint32_t error;

    bytes=packet.size+sizeof(packet_size_t);
    buffer=packet.u8;

    packet.seq=m_packet_seq.fetch_add(1);

    if(o)
    {
        o->m_event.reset();

        m_pending_requests_cs.lock();

        m_pending_requests[packet.seq]=o;

        m_pending_requests_cs.unlock();
    }

    m_send_cs.lock();

    while(bytes)
    {
        result=::send(m_socket,(char*)buffer,bytes,0);

        if(result<0)
        {
            error=errno;

            m_send_cs.unlock();

            if(o)
            {
                o->m_error=error;
                m_pending_requests_cs.lock();
                m_pending_requests.erase(packet.seq);
                m_pending_requests_cs.unlock();
                o->m_event.set();
            }
            return false;
        }

        bytes-=result;
        buffer=(char*)buffer+result;
    }

    m_send_cs.unlock();

    return true;
}

void rx_client::process_packet_ack(union tcp_packet *packet)
{
    m_pending_requests_cs.lock();

    auto it=m_pending_requests.find(packet->s_ack.seq);

    if(it!=m_pending_requests.end())
    {
        it->second->m_error=packet->s_ack.error;
        it->second->m_event.set();
    }

    m_pending_requests_cs.unlock();
}

void rx_client::process_packet_device_info(union tcp_packet *packet)
{
    packet->s_device_info.serial_number[sizeof(packet->s_device_info.serial_number)-1]=0;

    m_cs.lock();

    m_serial_number=packet->s_device_info.serial_number;

    m_cs.unlock();

    m_serial_number_changed=true;
}


void rx_client::process_packet_ddc1_samples(union tcp_packet *packet)
{
 float *buffer;
 uint32_t i;
 union
 {
     uint8_t u8[4];
     int32_t i32;
 } sample32;

    assert(packet->s_ddc1_samples.count==m_ddc1_sample_sets_to_send);


    m_ddc1_buffers_cs.lock();

    if(m_ddc1_buffers.size()>=MAX_DDC1_BUFFER_COUNT)
    {
        buffer=m_ddc1_buffers.front();
        m_ddc1_buffers.pop_front();
        m_ddc1_buffers_index+=m_ddc1_sample_sets_per_buffer;
    }
    else
    {
        buffer=(float*)malloc(m_ddc1_buffer_size);
    }

    m_ddc1_buffers_cs.unlock();


    if(packet->s_ddc1_samples.bits_per_sample==16)
    {
        int16_t *src=(int16_t*)packet->s_ddc1_samples.data;

        for(i=0;i<m_ddc1_sample_sets_to_send;i++)
        {
            buffer[i*2]=src[i*2]*(1.0f/32768.0f);
            buffer[i*2+1]=src[i*2+1]*(1.0f/32768.0f);
        }
    }
    else
    {
        uint8_t *src=packet->s_ddc1_samples.data;

        sample32.u8[0]=0;

        for(i=0;i<m_ddc1_sample_sets_to_send;i++)
        {
            sample32.u8[1]=src[i*6+0];
            sample32.u8[2]=src[i*6+1];
            sample32.u8[3]=src[i*6+2];

            buffer[i*2]=sample32.i32*(1.0f/2147483648.0f);

            sample32.u8[1]=src[i*6+3];
            sample32.u8[2]=src[i*6+4];
            sample32.u8[3]=src[i*6+5];

            buffer[i*2+1]=sample32.i32*(1.0f/2147483648.0f);
        }
    }

    m_ddc1_buffers_cs.lock();

    m_ddc1_buffers.push_back(buffer);

    m_ddc1_buffers_cs.unlock();

    if(m_callback)
    {
        m_callback->rx_client_callback_ddc1_samples(this,packet->s_ddc1_samples.index,buffer,m_ddc1_sample_sets_to_send,packet->s_ddc1_samples.bits_per_sample);
    }
}

void rx_client::process_packet_state(union tcp_packet *packet)
{
    m_cs.lock();

    m_device_state=packet->s_state.device;
    m_1pps_event_counter=packet->s_state.event_counter;
    m_1pps_adc_period_counter=packet->s_state.adc_period_counter;
    m_1pps_ddc1_sample_counter=packet->s_state.sample_counter;
    m_1pps_ddc1_adc_counter=packet->s_state.adc_counter;

    m_cs.unlock();
}

void rx_client::process_packet(union tcp_packet *packet)
{
    switch(packet->code)
    {
        case PACKET_CODE_S_DDC1_SAMPLES:
            process_packet_ddc1_samples(packet);
            break;

        case PACKET_CODE_S_STATE:
            process_packet_state(packet);
            break;

        case PACKET_CODE_S_ACK:
            process_packet_ack(packet);
            break;

        case PACKET_CODE_S_DEVICE_INFO:
            process_packet_device_info(packet);
            break;

        default:
            break;
    }
}

bool rx_client::send_extended_configuration_params(void)
{
 union tcp_packet packet;

    packet.size=PACKET_SIZE(c_extended_configuration_params);
    packet.code=PACKET_CODE_C_SET_EXTENDED_CONFIGURATION_PARAMS;
    packet.c_extended_configuration_params.index=m_index;
    packet.c_extended_configuration_params.device_count=m_device_count;
    packet.c_extended_configuration_params.delay=m_delay;

    return send(packet);
}


void rx_client::set_power(uint8_t power)
{
 union tcp_packet packet;

    power=power?1:0;

    if(m_power!=power)
    {
        m_power=power;

        packet.size=PACKET_SIZE(c_power);
        packet.code=PACKET_CODE_C_SET_POWER;
        packet.c_power.power=power;

        send(packet);
    }
}

bool rx_client::start_ddc1(uint32_t type_index,uint32_t sample_sets_per_buffer,uint32_t sample_sets_to_send)
{
 union tcp_packet packet;
 overlapped o;

    packet.code=PACKET_CODE_C_START_DDC1;
    packet.size=PACKET_SIZE(c_start_ddc1);
    packet.c_start_ddc1.type_index=type_index;
    packet.c_start_ddc1.sample_sets_per_buffer=sample_sets_per_buffer;
    m_ddc1_sample_sets_to_send=
    packet.c_start_ddc1.sample_sets_to_send=sample_sets_to_send;

    m_ddc1_buffers_index=0;

    m_ddc1_info=ddc_infos[type_index];
    m_ddc1_sample_sets_per_buffer=sample_sets_per_buffer;

    m_ddc1_buffer_size=sizeof(float)*2*sample_sets_to_send;

    send(packet,&o);

    if(o.m_event.wait(TIMEOUT))
    {
        return o.m_error==0;
    }

    return false;
}

bool rx_client::set_ddc1_frequency(int32_t frequency)
{
 union tcp_packet packet;
 overlapped o;

    packet.code=PACKET_CODE_C_SET_DDC1_FREQUENCY;
    packet.size=PACKET_SIZE(c_ddc1_frequency);
    packet.c_ddc1_frequency.frequency=frequency;

    send(packet,&o);

    if(o.m_event.wait(TIMEOUT))
    {
        return o.m_error==0;
    }

    return false;
}

void rx_client::set_ddc1_phase_shift(double phase_shift)
{
 union tcp_packet packet;

    packet.code=PACKET_CODE_C_SET_DDC1_PHASE_SHIFT;
    packet.size=PACKET_SIZE(c_ddc1_phase_shift);
    packet.c_ddc1_phase_shift.phase_shift=phase_shift;

    send(packet);
}

void rx_client::set_atten(uint32_t atten)
{
 union tcp_packet packet;

    packet.code=PACKET_CODE_C_SET_ATTEN;
    packet.size=PACKET_SIZE(c_atten);
    packet.c_atten.atten=atten;

    send(packet);
}

void rx_client::set_preamp(uint8_t preamp)
{
 union tcp_packet packet;

    packet.code=PACKET_CODE_C_SET_PREAMP;
    packet.size=PACKET_SIZE(c_preamp);
    packet.c_preamp.preamp=preamp;

    send(packet);
}

void rx_client::set_1pps_enabled(uint8_t enabled)
{
 union tcp_packet packet;

    packet.code=PACKET_CODE_C_SET_1PPS;
    packet.size=PACKET_SIZE(c_1pps);
    packet.c_1pps.enabled=enabled;

    send(packet);
}


uint32_t rx_client::trigger(void)
{
 union tcp_packet packet;
 overlapped o;

    packet.code=PACKET_CODE_C_TRIGGER;
    packet.size=EMPTY_PACKET_SIZE;

    send(packet,&o);

    if(!o.m_event.wait(TIMEOUT))
    {
        return ETIMEDOUT;
    }

    return o.get_error();
}

uint32_t rx_client::trigger_test(void)
{
 union tcp_packet packet;
 overlapped o;

    packet.code=PACKET_CODE_C_TRIGGER_TEST;
    packet.size=EMPTY_PACKET_SIZE;

    send(packet,&o);

    if(!o.m_event.wait(TIMEOUT))
    {
        return ETIMEDOUT;
    }

    return o.get_error();
}

void rx_client::free_buffers(void)
{
    while(!m_ddc1_buffers.empty())
    {
        free(m_ddc1_buffers.front());
        m_ddc1_buffers.pop_front();
    }
}

void rx_client::stop_ddc1(void)
{
 union tcp_packet packet;
 overlapped o;

    packet.code=PACKET_CODE_C_STOP_DDC1;
    packet.size=EMPTY_PACKET_SIZE;

    send(packet,&o);

    o.m_event.wait(TIMEOUT);

    free_buffers();
}

bool rx_client::get_buffer(uint64_t index,float *buffer)
{
 uint32_t buffer_index;


    m_ddc1_buffers_cs.lock();

    buffer_index=(uint32_t)((index-m_ddc1_buffers_index)/m_ddc1_sample_sets_per_buffer);

    if(buffer_index<m_ddc1_buffers.size())
    {
        memcpy(buffer,m_ddc1_buffers[buffer_index],m_ddc1_buffer_size);

        m_ddc1_buffers_cs.unlock();
        return true;
    }

    m_ddc1_buffers_cs.unlock();

    return false;
}

std::string rx_client::get_remote_host_desc(void)
{
 int len;
 char s[512];
 std::string sn;

    sn=get_serial_number();

    if(!sn.empty())
    {
        len=sprintf(s,"%s - ",sn.c_str());
    }
    else
    {
        len=0;
    }

    if(m_host_name.empty())
    {
        sprintf(s+len,"%s:%u",m_address_str.c_str(),m_port);
    }
    else
    {
        sprintf(s+len,"%s:%u",m_host_name.c_str(),m_port);
    }

    return s;
}
