/*****************************************************************************/
/*                      WiNRADiO G65DDC EXCALIBUR SIGMA                      */
/*                       Coherent mode SDK DEMO - ETC                        */
/*                        (c) Copyright RADIXON 2023                         */
/*****************************************************************************/

#ifndef __U_RX_CLIENT_H__
#define __U_RX_CLIENT_H__

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include "../u_packet.h"
#include "u_wx.h"
#include <map>
#include <atomic>
#include "u_ddc_info.h"
#include <deque>
#include <mutex>
#include <condition_variable>
#include <thread>

class remote_host
{
public:
    remote_host(uint32_t index,int32_t delay,const std::string &host_name,uint16_t port)
    {
        m_host_name=host_name;
        m_port=port;
        m_delay=delay;
        m_index=index;
    }

    std::string m_host_name;
    uint16_t    m_port;
    int32_t     m_delay;
    uint32_t    m_index;
};

class event
{
public:
    void set(void)
    {
        m_mutex.lock();
        m_state=true;
        m_mutex.unlock();
        m_cond.notify_all();
    }

    void reset(void)
    {
        m_mutex.lock();
        m_state=false;
        m_mutex.unlock();
    }

    void wait(void)
    {
        std::unique_lock<decltype(m_mutex)> lock(m_mutex);
        m_cond.wait(lock,[&]{return m_state;});
    }

    bool wait(uint32_t timeout_ms)
    {
        std::unique_lock<decltype(m_mutex)> lock(m_mutex);
        return m_cond.wait_for(lock,std::chrono::milliseconds(timeout_ms),[&]{return m_state;});
    }

    event(bool state=false):m_state(state)
    {
    }
private:
    bool                        m_state;
    std::mutex                  m_mutex;
    std::condition_variable     m_cond;
};

class overlapped
{
public:
    overlapped(void)
    {
        m_error=0;
    }

    uint32_t get_error(void) const
    {
        return m_error;
    }
private:
    friend class rx_client;
    event       m_event;
    uint32_t    m_error;
};


class rx_client;

class rx_client_callback
{
public:
    virtual void rx_client_callback_ddc1_samples(rx_client *client,uint64_t index,float *samples,uint32_t count,uint32_t bits_per_sample)=0;
};

class rx_client
{
public:
    int connect(void);
    void disconnect(void);

    bool get_buffer(uint64_t index,float *buffer);

    void set_power(uint8_t power);

    bool start_ddc1(uint32_t type_index,uint32_t sample_sets_per_buffer,uint32_t sample_sets_to_send);
    void stop_ddc1(void);

    bool set_ddc1_frequency(int32_t frequency);
    void set_ddc1_phase_shift(double phase_shift);

    void set_preamp(uint8_t preamp);
    void set_atten(uint32_t atten);

    void set_1pps_enabled(uint8_t enabled);

    uint32_t trigger(void);
    uint32_t trigger_test(void);

    const std::string& get_address_str(void) const
    {
        return m_address_str;
    }

    uint16_t get_port(void) const
    {
        return m_port;
    }

    std::string get_serial_number(void)
    {
        std::lock_guard<decltype(m_cs)> lock(m_cs);
        return m_serial_number;
    }

    void get_state(G65DDC_DEVICE_STATE *device_state,uint32_t *event_counter,uint32_t *adc_period_counter,double *ddc1_sample_counter,uint64_t *ddc1_adc_counter)
    {
        m_cs.lock();
        *device_state=m_device_state;
        *event_counter=m_1pps_event_counter;
        *adc_period_counter=m_1pps_adc_period_counter;
        *ddc1_sample_counter=m_1pps_ddc1_sample_counter;
        *ddc1_adc_counter=m_1pps_ddc1_adc_counter;
        m_cs.unlock();
    }

    bool get_serial_number_changed(void)
    {
        return m_serial_number_changed.exchange(false);
    }

    const std::string& get_host_name(void) const
    {
        return m_host_name;
    }

    std::string get_remote_host_desc(void);

    uint32_t get_index(void) const
    {
        return m_index;
    }

    int32_t get_delay(void) const
    {
        return m_delay;
    }

    void set_callback(rx_client_callback *callback)
    {
        m_callback=callback;
    }

    rx_client(uint32_t index,uint32_t device_count,int32_t delay,const std::string &host_name,struct in_addr addr,uint16_t port);
    ~rx_client(void);
private:
    int                                 m_socket;
    struct in_addr                      m_addr;
    std::string                         m_address_str;
    uint16_t                            m_port;
    std::string                         m_serial_number;
    std::string                         m_host_name;
    std::thread                         *m_thread;
    bool                                m_stop_request;
    bool                                m_connected;
    std::map<packet_seq_t,overlapped*>  m_pending_requests;
    std::recursive_mutex                m_pending_requests_cs;
    std::atomic<packet_seq_t>           m_packet_seq;
    std::recursive_mutex                m_send_cs;
    std::recursive_mutex                m_cs;
    std::atomic<bool>                   m_serial_number_changed;

    uint32_t                            m_index;
    uint32_t                            m_device_count;
    int32_t                             m_delay;
    uint8_t                             m_power;
    struct ddc_info                     m_ddc1_info;
    uint32_t                            m_ddc1_sample_sets_per_buffer;
    uint32_t                            m_ddc1_sample_sets_to_send;
    uint32_t                            m_ddc1_buffer_size;

    G65DDC_DEVICE_STATE                 m_device_state;
    uint32_t                            m_1pps_event_counter;
    uint32_t                            m_1pps_adc_period_counter;
    double                              m_1pps_ddc1_sample_counter;
    uint64_t                            m_1pps_ddc1_adc_counter;

    rx_client_callback                  *m_callback;

    std::recursive_mutex                m_ddc1_buffers_cs;
    std::deque<float*>                  m_ddc1_buffers;
    uint64_t                            m_ddc1_buffers_index;

    inline packet_seq_t get_packet_seq(void)
    {
        return m_packet_seq.fetch_add(1);
    }

    void thread_procedure(void);

    void free_buffers(void);

    bool send(union tcp_packet &packet,overlapped *ovrlp=NULL);

    bool send_extended_configuration_params(void);

    void process_packet(union tcp_packet *packet);
    void process_packet_ack(union tcp_packet *packet);
    void process_packet_device_info(union tcp_packet *packet);
    void process_packet_ddc1_samples(union tcp_packet *packet);
    void process_packet_state(union tcp_packet *packet);
};

#endif //__U_RX_CLIENT_H__

